from engineio_client.emitter import Emitter
from engineio_client.utils import format_long
from .parser import Packet

import logging
logger = logging.getLogger(__name__)

class Socket(object):
    def __init__(self, namespace, manager):
        self.namespace = namespace
        self.manager = manager

        self.state = 'disconnected'
        self.event_handlers = Emitter()
        self.socket_handlers = Emitter()
        self.acks = {}
        self.ids = 0

        if self.manager.auto_connect:
            self.connect()

    def setup_manager_handlers(self):
        self.manager.attach_socket(self)
        self.manager.on('open', self.handle_open)
        self.manager.on('close', self.handle_close)
        self.manager.on('packet', self.handle_packet)

    def cleanup_manager_handlers(self):
        self.manager.off('open', self.handle_open)
        self.manager.off('close', self.handle_close)
        self.manager.off('packet', self.handle_packet)
        self.manager.detach_socket(self)
    
    def cleanup_handlers(self):
        self.socket_handlers.removeAllListeners()
        self.event_handlers.removeAllListeners()

    @property
    def id(self):
        return self.manager.id

    def connect(self):
        if self.state in ['connecting', 'connected']:
            return

        self.state = 'connecting'
        self.manager.connect()   # Ensure connected
        if self.manager.state == 'open':
            self.handle_open()
        self.socket_handlers.emit('connecting')
        self.setup_manager_handlers()

    def disconnect(self):
        if self.state not in ['connecting', 'connected']:
            return

        logger.debug("Sending disconnect")
        self.send_packet(Packet(Packet.DISCONNECT))
        self.cleanup_manager_handlers()
        self.handle_close()
        self.cleanup_handlers()

    def on(self, *args, **kwargs):
        """Attach handler for event-type messages"""
        return self.event_handlers.on(*args, **kwargs)

    def on_socket(self, *args, **kwargs):
        """Attach handler for generic socket events"""
        return self.socket_handlers.on(*args, **kwargs)

    def on_connecting(self, *args, **kwargs):
        return self.on_socket('connecting', *args, **kwargs)

    def on_connect(self, *args, **kwargs):
        return self.on_socket('connect', *args, **kwargs)

    def on_disconnect(self, *args, **kwargs):
        return self.on_socket('disconnect', *args, **kwargs)

    def on_error(self, *args, **kwargs):
        return self.on_socket('error', *args, **kwargs)

    def on_event(self, *args, **kwargs):
        return self.on_socket('event', *args, **kwargs)

    def emit(self, *args, **kwargs):
        """Send event-type message
        If callback is provided, it will be called when the server want to
        respond to the event (maybe never).
        """
        logger.debug(format_long("Sending event: %s", args))

        packet = Packet(type=Packet.EVENT, data=args)

        callback = kwargs.get('callback', None)
        if callback:
            packet.id = self.ids
            self.acks[packet.id] = callback
            self.ids += 1

        self.send_packet(packet)

    def send_packet(self, packet):
        packet.namespace = self.namespace
        self.manager.send_packet(packet)

    def handle_open(self):
        logger.debug("Manager is open, connecting")
        if self.namespace != '/':
            self.send_packet(Packet(Packet.CONNECT))

    def handle_close(self):
        logger.debug("Closed")
        self.state = 'disconnected'
        self.socket_handlers.emit('disconnect')

    def handle_packet(self, packet):
        if packet.namespace != self.namespace:
            return

        if packet.type == Packet.CONNECT:
            self.handle_connect()
        elif packet.type == Packet.DISCONNECT:
            self.handle_disconnect()
        elif packet.type in [Packet.EVENT, Packet.BINARY_EVENT]:
            self.handle_event(packet)
        elif packet.type in [Packet.ACK, Packet.BINARY_ACK]:
            self.handle_ack(packet)
        elif packet.type == Packet.ERROR:
            self.handle_error(packet.data)
        else:
            logger.error("Invalid packet type: %d", packet.type)

    def handle_connect(self):
        logger.debug("Received connect")
        self.state = 'connected'
        self.socket_handlers.emit('connect')

    def handle_disconnect(self):
        logger.debug("Received disconnect")
        self.cleanup_manager_handlers()
        self.handle_close()
        self.cleanup_handlers()

    def handle_event(self, packet):
        args = packet.data or []
        logger.debug(format_long("Received event: %s", args))

        kwargs = {}
        if packet.id:
            kwargs['callback'] = self.create_ack_callback(packet.id)

        self.event_handlers.emit(*args, **kwargs)
        self.socket_handlers.emit('event', *args, **kwargs)

    def handle_ack(self, packet):
        args = packet.data or []
        logger.debug(format_long("Received ack: %s", args))
        
        try:
            callback = self.acks.pop(packet.id)
        except KeyError:
            logger.warning("Invalid ack id")
        else:
            callback(*args)

    def handle_error(self, error):
        logger.warning("Received error: %s", error)
        self.socket_handlers.emit('error', error)

    def create_ack_callback(self, id):
        def callback(*args):
            if callback.sent:
                logger.warning("Ack already sent")
                return

            callback.sent = True
            logger.debug(format_long("Sending ack: %s", args))
            self.send_packet(Packet(type=Packet.ACK, id=id, data=args))
        callback.sent = False

        return callback




from __future__ import absolute_import
import queue
import threading
import time

try:
    from simple_websocket import Server, ConnectionClosed
    _websocket_available = True
except ImportError:  # pragma: no cover
    _websocket_available = False


class WebSocketWSGI(object):  # pragma: no cover
    """
    This wrapper class provides a threading WebSocket interface that is
    compatible with eventlet's implementation.
    """
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        self.ws = Server(environ)
        return self.app(self)

    def close(self):
        return self.ws.close()

    def send(self, message):
        try:
            return self.ws.send(message)
        except ConnectionClosed:
            raise IOError()

    def wait(self):
        try:
            return self.ws.receive()
        except ConnectionClosed:
            raise IOError()


_async = {
    'thread': threading.Thread,
    'queue': queue.Queue,
    'queue_empty': queue.Empty,
    'event': threading.Event,
    'websocket': WebSocketWSGI if _websocket_available else None,
    'sleep': time.sleep,
}



from ..transport import Transport
from ..parser import Packet
from ..utils import format_long
import requests
import gevent
import json

import logging
logger = logging.getLogger(__name__)

class Polling(Transport):
    name = 'polling'

    def __init__(self, *args, **kwargs):
        super(Polling, self).__init__(*args, **kwargs)
        self.session = requests.Session()
        self.session.headers.update({'Content-Type': 'application/octet-stream'})
        self.reading = False
        self.writing = False
        self.read_loop = None

    def get_uri(self):
        query = {
            'EIO': '3',
            'transport': self.name
        }
        if self.client and self.client.sid:
            query['sid'] = self.client.sid
        querystring = '?' + '&'.join(['='.join(item) for item in query.iteritems()]) if query else ''

        return '%s://%s:%d%s/%s' % (self.scheme, self.hostname, self.port, self.path, querystring)

    def handle_payload(self, payload):
        packets = self.parser.decode_payload(payload)
        for packet in packets:
            if self.state == 'opening':
                self.handle_open()
            self.handle_packet(packet)

    def loop_read(self):
        while self.state in ['opening', 'open']:
            try:
                payload = self.read()
                self.handle_payload(payload)
            except requests.RequestException as e:
                self.handle_error(e)

    def do_open(self):
        self.read_loop = self.client.start_loop(self.loop_read)

    def do_close(self):
        def close():
            logger.debug('Writing close packet')
            self.send([Packet(Packet.CLOSE)])
            self.client.stop_loop(self.read_loop)

        if self.state == 'open':
            logger.debug('Closing')
            close()
        else:
            logger.debug('Deferring close');
            self.once('open', close)

    def do_send(self, packets):
          payload = self.parser.encode_payload(packets)
          try:
              self.write(payload)
          except requests.RequestException as e:
              self.handle_error(e)

    def handle_pause(self):
        self.state = 'paused'
        self.emit('pause')

    def pause(self):
        self.state = 'pausing'

        if not self.reading and not self.writing:
            self.handle_pause()
        
        remaining_tasks = 0
        def terminate_task():
            remaining_tasks -= 1
            if remaining_tasks <= 0:
                self.handle_pause()
        
        if self.reading:
            remaining_tasks += 1
            self.once('read-done', terminate_task)

        if self.writing:
            remaining_tasks += 1
            self.once('write-done', terminate_task)

    def read(self):
        self.reading = True
        logger.debug("Polling")
        r = self.session.get(self.get_uri(), stream=True)

        self.reading = False
        self.emit('read-done')

        r.raise_for_status()
        payload = r.raw.read()
        logger.debug(format_long("Received payload: %s", repr(payload)))
        return payload

    def write(self, payload):
        self.writing = True
        logger.debug(format_long("Sending payload: %s", repr(payload)))
        r = self.session.post(self.get_uri(), stream=True, data=payload)

        self.writing = False
        self.emit('write-done')

        r.raise_for_status()




"""Contains wrappers for builtins.

Each wrapper is used to avoid context manager boilerplate.

Attributes:
    read_json: Wrapped ``json.load``.
    write_json: Wrapped ``json.dump``.
    read_pickle: Wrapped ``pickle.load``.
    write_pickle: Wrapped ``pickle.dump``.

"""

import json
import pickle

from .utils import read_wrapper, write_wrapper


def read(file, mode='r', **kwargs):
    """Read and return content of the file.

    Args:
        file: Path-like object giving the pathname of the file to be opened.
        mode: String giving the mode of the file to be opened. Default: 'r'.
        **kwargs: Optional arguments that ``open`` takes.

    Returns:
        File content.

    """
    with open(file, mode, **kwargs) as f:
        return f.read()


def write(file, s, mode='w', **kwargs):
    """Read and return content of the file.

    Args:
        file: Path-like object giving the pathname of the file to be opened.
        s: Content to be written.
        mode: String giving the mode of the file to be opened. Default: 'w'.
        **kwargs: Optional arguments that ``open`` takes.

    Returns:
        The number of characters/bytes (according to mode) written.

    """
    with open(file, mode, **kwargs) as f:
        return f.write(s)


read_json = read_wrapper(json.load, mode='r')
write_json = write_wrapper(json.dump, mode='w')

read_pickle = read_wrapper(pickle.load, mode='rb')
write_pickle = write_wrapper(pickle.dump, mode='wb')

__all__ = ['read', 'write',
           'read_json', 'write_json',
           'read_pickle', 'write_pickle']



import re

import shortio

from .__version__ import __version__


VOWELS = re.compile('[aeiou]')

__all__ = []
for name in shortio.__all__:
    name_parts = name.split('_')

    if len(name_parts) > 1:
        prefix = ''.join([part[0] for part in name_parts[:-1]])
        frmt = VOWELS.sub('', name_parts[-1])
        new_name = f'{prefix}{frmt}'
    else:
        new_name = name_parts[0][0]

    globals()[new_name] = getattr(shortio, name)
    __all__.append(new_name)
